# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from glob import glob

from datalad_crawler.pipelines.tests.test_utils import _test_smoke_pipelines as _tsp
from datalad.utils import chpwd
from datalad.utils import _path_
from datalad.tests.utils_pytest import eq_
from datalad.tests.utils_pytest import assert_false
from datalad.tests.utils_pytest import with_tempfile
from datalad.tests.utils_pytest import use_cassette
from datalad.tests.utils_pytest import externals_use_cassette
from datalad.tests.utils_pytest import skip_if_no_network
from datalad.tests.utils_pytest import ok_clean_git
from datalad.tests.utils_pytest import ok_file_under_git
from ..simple_s3 import pipeline
from datalad.api import crawl_init
from datalad.api import crawl
from datalad.api import create
from datalad.support.annexrepo import AnnexRepo
from datalad.downloaders.tests.utils import get_test_providers

import pytest

from logging import getLogger
lgr = getLogger('datalad.crawl.tests')


@pytest.mark.parametrize("func,args,kwargs", [
    (pipeline, ["b"], {}),
    # to_http everywhere just to make it faster by avoiding initiating datalad
    # special remote
    (pipeline, ["b"], dict(to_http=True, prefix="prefix")),
    (pipeline, ["b"], dict(to_http=True)),
    (pipeline, ["b"], dict(to_http=True, archive=True)),
    (pipeline, ["b"], dict(to_http=True, directory="subdataset", prefix="some/")),
])
def test_smoke_pipelines(func, args, kwargs):
    _tsp(pipeline, args, kwargs)

@with_tempfile
@skip_if_no_network
def _test_drop(path=None, *, drop_immediately):
    s3url = 's3://datalad-test0-nonversioned'
    providers = get_test_providers(s3url)  # to verify having s3 credentials
    # vcr tape is getting bound to the session object, so we need to
    # force re-establishing the session for the bucket.
    # TODO (in datalad): make a dedicated API for that, now too obscure
    _ = providers.get_status(s3url, allow_old_session=False)
    create(path)
    # unfortunately this doesn't work without force dropping since I guess vcr
    # stops and then gets queried again for the same tape while testing for
    # drop :-/
    with chpwd(path):
        crawl_init(
            template="simple_s3",
            args=dict(
                bucket="datalad-test0-nonversioned",
                drop=True,
                drop_force=True,  # so test goes faster
                drop_immediately=drop_immediately,
            ),
            save=True
        )
    if drop_immediately:
        # cannot figure out but taping that interaction results in
        # git annex addurl  error.  No time to figure it out
        # so we just crawl without vcr for now. TODO: figure out WTF
        with chpwd(path):
            crawl()
    else:
        with externals_use_cassette(
                'test_simple_s3_test0_nonversioned_crawl_ext'
                + ('_immediately' if drop_immediately else '')), \
                chpwd(path):
            crawl()
    # test that all was dropped
    repo = AnnexRepo(path, create=False)
    files = glob(_path_(path, '*'))
    eq_(len(files), 8)
    for f in files:
        assert_false(repo.file_has_content(f))


@use_cassette('test_simple_s3_test0_nonversioned_crawl')
def test_drop():
    _test_drop(drop_immediately=False)


@use_cassette('test_simple_s3_test0_nonversioned_crawl_immediately')
def test_drop_immediately():
    _test_drop(drop_immediately=True)


@with_tempfile
@use_cassette('test_simple_s3_test2_obscurenames_versioned_crawl')
@skip_if_no_network
def test_obscure_names(path=None):
    bucket = "datalad-test2-obscurenames-versioned"
    get_test_providers('s3://' + bucket)  # to verify having s3 credentials
    create(path)
    with externals_use_cassette('test_simple_s3_test2_obscurenames_versioned_crawl_ext'), \
         chpwd(path):
        crawl_init(template="simple_s3",
                   args=dict(bucket=bucket),
                   save=True
                   )
        crawl()
    # fun with unicode was postponed
    ok_clean_git(path, annex=True)
    for f in [
        'f &$=@:+,?;', "f!-_.*'( )", 'f 1', 'f [1][2]'
    ]:
        ok_file_under_git(path, f, annexed=True)



"""A pipeline for Stanford Repository's xml datasets"""

from ..nodes.matches import xpath_match
from ..nodes.crawl_url import crawl_url
try:
    from scrapy.http import Response
    from scrapy.http import XmlResponse
    from scrapy.selector import Selector

except ImportError:  # pragma: no cover
    class Selector(object):
        xpath = css = None
    Response = None


# copied the simple_with_archives template in order to show
def pipeline(doc_id=None,
             x_pathmatch_="//file/@id",
             leading_dirs_depth=1):
    """Pipeline to crawl xml datasets from purls
    """

    crawler = crawl_url("https://purl.stanford.edu/" + doc_id + ".xml")
    # adding print_xml to incoming pipeline
    incoming_pipeline = [  # Download all the archives found on the project page
        crawler,
        xpath_match(x_pathmatch_, min_count=1), # changed h-ref to xpath_match
        configure_url,
        print_xml
    ]

    from .simple_with_archives import pipeline as sa_pipeline
    return sa_pipeline(leading_dirs_depth=leading_dirs_depth,
                       add_annex_to_incoming_pipeline=True,
                       incoming_pipeline=incoming_pipeline)


def configure_url(data, keys=['url'], match='match'):
    """Given purl, configure to a url that leads directly to the file
    """
    data = data.copy()

    for key in keys:
        url_elements = data[key].split("/")
        doc_id = url_elements[3][:-4]
        data[key] = "https://stacks.stanford.edu/file/druid:" + doc_id + "/" + data[match]
    yield data


# print generator
def print_xml(data, keys=['url']):
    """Given xml data, get value within 'xml' key and print it
    """
    data = data.copy()
    for key in keys:
        print(key, (data[key]))
    yield data



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##


from datalad.tests.utils_pytest import known_failure_direct_mode, eq_, assert_raises, assert_in
from mock import patch
from datalad.api import crawl_init
from collections import OrderedDict
from os.path import exists
from datalad.support.annexrepo import AnnexRepo
from datalad.tests.utils_pytest import with_tempfile, chpwd
from datalad.tests.utils_pytest import ok_clean_git
from datalad_crawler.consts import CRAWLER_META_CONFIG_PATH, CRAWLER_META_DIR
from datalad.distribution.dataset import Dataset

import pytest


@pytest.mark.parametrize("args,template,template_func,save,target_value", [
    (None, 'openfmri', 'superdataset_pipeline', False,
          '[crawl:pipeline]\ntemplate = openfmri\nfunc = superdataset_pipeline\n\n'),
    ({'dataset': 'ds000001'}, 'openfmri', None, False,
          '[crawl:pipeline]\ntemplate = openfmri\n_dataset = ds000001\n\n'),
    (['dataset=ds000001', 'versioned_urls=True'], 'openfmri', None, False,
          '[crawl:pipeline]\ntemplate = openfmri\n_dataset = ds000001\n_versioned_urls = True\n\n'),
    (None, 'openfmri', 'superdataset_pipeline', True,
          '[crawl:pipeline]\ntemplate = openfmri\nfunc = superdataset_pipeline\n\n'),
])
@with_tempfile(mkdir=True)
def test_crawl_init(tmpdir=None, *, args, template, template_func, save, target_value):
    ar = AnnexRepo(tmpdir, create=True)
    with chpwd(tmpdir):
        crawl_init(args=args, template=template, template_func=template_func, save=save)
        eq_(exists(CRAWLER_META_DIR), True)
        eq_(exists(CRAWLER_META_CONFIG_PATH), True)
        f = open(CRAWLER_META_CONFIG_PATH, 'r')
        contents = f.read()
        eq_(contents, target_value)
        if save:
            ds = Dataset(tmpdir)
            ok_clean_git(tmpdir, annex=isinstance(ds.repo, AnnexRepo))


@pytest.mark.parametrize("args,template,template_func,target_value", [
    ('tmpdir', None, None, ValueError),
    (['dataset=Baltimore', 'pie=True'], 'openfmri', None, RuntimeError),
    (None, None, None, TypeError),
])
@with_tempfile(mkdir=True)
def test_crawl_init_error(tmpdir=None, *, args, template, template_func, target_value):
        ar = AnnexRepo(tmpdir)
        with chpwd(tmpdir):
            assert_raises(target_value, crawl_init, args=args, template=template, template_func=template_func)


@pytest.mark.parametrize("return_value,exc,exc_msg", [
    ([], ValueError, "returned pipeline is empty"),
    ({1: 2}, ValueError, "pipeline should be represented as a list. Got: {1: 2}"),
])
@with_tempfile(mkdir=True)
def test_crawl_init_error_patch(d=None, *, return_value, exc, exc_msg):
    ar = AnnexRepo(d, create=True)
    with patch('datalad_crawler.crawl_init.load_pipeline_from_template',
               return_value=lambda dataset: return_value) as cm:
        with chpwd(d):
            with assert_raises(exc) as cm2:
                crawl_init(args=['dataset=Baltimore'], template='openfmri')
            assert_in(exc_msg, str(cm2.value))

            cm.assert_called_with('openfmri', None, return_only=True, kwargs=OrderedDict([('dataset', 'Baltimore')]))



from datalad.utils import chpwd

from datalad.api import (
    crawl,
    crawl_init,
    create,
)
try:
    from datalad.support.github_ import _get_github_cred
except ImportError:
    # might be dated which has not merged
    # https://github.com/datalad/datalad/pull/4400 yet
    from datalad.downloaders.credentials import UserPassword
    def _get_github_cred():
        """Trimmed down helper"""
        return UserPassword("github", "does not matter")

from datalad.tests.utils_pytest import (
    assert_false,
    assert_greater,
    skip_if_no_network,
    with_tempfile,
)
import pytest


@skip_if_no_network
@with_tempfile
def test_crawl(tempd=None):
    if not _get_github_cred().is_known:
        pytest.skip("no github credential")
    ds = create(tempd)
    with chpwd(tempd):
        crawl_init(
            template='gh', save=True,
            args={'org': 'datalad-collection-1', 'include': 'kaggle'}
        )
        crawl()
    subdss = ds.subdatasets(fulfilled=True, result_xfm='datasets')
    assert all('kaggle' in d.path for d in subdss)
    assert_greater(len(subdss), 1)
    assert_false(ds.repo.dirty)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""A pipeline for crawling an arbitrary GitHub organization"""

# I guess due to all the "import magic" we can't simply force load of
# global  github module whenever there is a local one, so we will rename
# local one

import re

from datalad.api import Dataset, install
from datalad.support import path as op
from datalad.support.gitrepo import GitRepo
from datalad.utils import (
    assure_bool,
    assure_list_from_str,
    rmtree,
    updated
)
from datalad.downloaders.credentials import UserPassword

# Possibly instantiate a logger if you would like to log
# during pipeline creation
from logging import getLogger
lgr = getLogger("datalad.crawler.pipelines.github")


def pipeline(org=None,
             repo_type='sources', include='.*', exclude=None,
             metadata_nativetypes=None, aggregate=False,
             get_data=False, drop_data=False):
    """Crawl github organization and install all matching repos as subdatasets

    Parameters
    ----------
    org: str
      GitHub Organization
    repo_type: ('all', 'public', 'private', 'forks', 'sources', 'member'), optional
      Type of the repositories to consider. Passed to github's module .get_repo.
    include: str, optional
      (search) regex to define which repositories to consider
    exclude: str, optional
      (search) regex to define which of the considered repositories to exclude.
      By default we exclude all repositories with '--' in the name which we
      assume defines the dataset boundary, so for `ds1--subds2` there should
      also be `ds` repository which would define the relationship to `subds2`
    metadata_nativetypes: list, optional
      List of metadata native types to define (locally within .git/config to
      not change repository itself) for metadata aggregation
    aggregate: bool, optional
      Aggregate metadata after crawling a new dataset
    get_data: bool, optional
      Get data for a new dataset
    drop_data: bool, optional
      Drop data for a new dataset after being done

    Returns
    -------
    list:
      pipeline
    """
    assert org, "Organization must be provided"
    aggregate = assure_bool(aggregate)
    get_data = assure_bool(get_data)
    drop_data = assure_bool(drop_data)

    import github as gh
    # TODO: consider elevating that function to a "public" helper
    from datalad.support.github_ import _gen_github_entity
    superds = Dataset('.')
    if metadata_nativetypes:
        metadata_nativetypes = assure_list_from_str(metadata_nativetypes, sep=',')

    aggregate_later = []
    def crawl_github_org(data):
        assert list(data) == ['datalad_stats'], data
        # TODO: actually populate the datalad_stats with # of datasets and
        # possibly amount of data downloaded in get below
        # Needs DataLad >= 0.13.6~7^2~3 where password was removed
        entity, cred = next(_gen_github_entity(None, org))
        all_repos = list(entity.get_repos(repo_type))

        for repo in all_repos:
            name = repo.name
            if include and not re.search(include, name):
                lgr.debug(
                    "Skipping %s since include regex search returns nothing", name
                )
                continue
            if exclude and re.search(exclude, name):
                lgr.debug(
                    "Skipping %s since exclude regex search returns smth", name
                )
                continue
            # Let's just do everything here
            dspath = name
            if op.exists(dspath):
                lgr.info("Skipping %s since already exists", name)
                # although we could just do install, which would at least
                # verify that url is the same... to not try to aggregate
                # etc, we will just skip for now
                continue

            # TODO: all the recursive etc options
            try:
                ds = superds.install(
                    dspath, source=repo.clone_url, get_data=get_data,
                    on_failure='continue'
                )
            except Exception as exc:
                if all(f.get('action', '') == 'add_submodule' and f.get('status', '') == 'error' for f in exc.failed):
                    # since we do not like nice exceptions and want to parse arbitrary text
                    # in the return records... let's resist that urge and redo the check
                    # since if no commit -- likely reason is an empty repo
                    if GitRepo(dspath).get_hexsha() is None:
                        lgr.warning(
                            "Cloned an empty repository.  Removing and proceeding without error"
                        )
                        rmtree(dspath)
                        continue
                if all(f.get('action', '') == 'get' for f in exc.failed):
                    lgr.warning(
                        "We failed to obtain %d files, extracted metadata etc might be incomplete",
                        len(exc.failed)
                    )
                    ds = Dataset(exc.failed[0]['refds'])
                else:
                    raise

            if metadata_nativetypes:
                lgr.info("Setting native metadata types to include %s",
                         ", ".join(metadata_nativetypes))
                nativetypes = ds.config.obtain('datalad.metadata.nativetype',
                                               default=[])
                for nativetype in metadata_nativetypes:
                    if nativetype not in nativetypes:
                        lgr.debug("Adding %s nativetype", nativetype)
                        ds.config.add('datalad.metadata.nativetype', nativetype, where='local')
                    else:
                        lgr.debug("Not adding %s nativetype since already defined", nativetype)
            # if anyone down the line needs it
            aggregate_later.append(dspath)
            yield {
                'dataset': ds,
                'superdataset': superds,
                'dataset_path': dspath,
                'dataset_name': name,
                'dataset_url': repo.clone_url,
            }


    def do_aggregate_metadata(data):
        # For now just aggregate into superdataset, and assume it is always provided!
        agg_out = data['superdataset'].aggregate_metadata(
            data['dataset_path'],
            recursive=True,  # I feel greedy
            incremental=True,
            update_mode='target'  # we aggregate into superdataset!
        )
        yield dict(data, aggregate_metadata_result=agg_out)

    def do_drop_data(data):
        lgr.info("Dropping all data for %(dataset)s", data)
        # This should fail to drop annexed metadata since no copies
        # would be known
        # Actually - no, since we aggregate only into its super!
        ds = data['dataset']
        drop_res = ds.drop('.')
        # notTODO: annex forget -- since initial crawl, should be perfectly
        # fine and might be of assistance to reduce impact on .git size
        # Nope -- we do not want to mess with original openneuro's git-annex
        # history.
        ds.repo.gc(allow_background=False)
        yield dict(data, drop_result=drop_res)

    def do_superds_aggregate_metadata(data):
        # nothing in data would be of use, and we know superds
        yield dict(
            data,
            aggregate_metadata_super_result=superds.aggregate_metadata(
                '.',
                incremental=True,
            )
        )

    # annex = Annexificator(no_annex=True, allow_dirty=True)
    lgr.info("Creating a pipeline for github organization %s", org)
    pipeline = [
            crawl_github_org,
    ]

    if aggregate:
        pipeline.append(do_aggregate_metadata)

    if drop_data:
        pipeline.append(do_drop_data)

    if aggregate:
        # but also we need a final aggregate for super ds itself
        pipeline = [
            pipeline,
            do_superds_aggregate_metadata
        ]

    return pipeline


# # There is nothing github'ish in it
# and actually no reason for some pipeline for now
# # We assume that repo already cloned, so "crawling" would only consists of
# # call to update
# def pipeline(merge=True, recursive=False, reobtain_data=False):
#     recursive = assure_bool(recursive)
#     merge = assure_bool(merge)
#     reobtain_data = assure_bool(reobtain_data)
#
#     def _update(data):
#         assert not data  # nothing is expected
#         for r in update(merge=merge, recursive=recursive, reobtain_data=reobtain_data):
#             yield updated(data, {'update_result': r})
#
#     return [
#         _update
#     ]



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Various utils which were moved from datalad core or might be later moved into it

To simplify such refactoring tests should be provided here as well
"""

import inspect


def flatten(l, types=None, base_type=None):
    """Given a list, return a recursively flattened version

    Parameters
    ----------
    l:
      Iterable to flatten
    types: type or tuple of types, optional
      By default all non-list types are preserved as is, but you could specify
      other iterable types (e.g., `set`, `tuple`, `dict`) which would be flattened
      then too.  Note that no (re)ordering of associative (`set`, `dict`)
      elements is done, thus order is not guaranteed if original container
      does not provide order
    base_type: type, optional
      What should be the output type. If not specified, the type of the `l`
      is used. Note that the function supports only the base_types which support
      `+` operand, so cannot be `set` or `dict`
    """
    if base_type is None:
        base_type = type(l)
    if types is None:
        types = base_type
    return sum((flatten(i, types=types, base_type=base_type) for i in l),
               base_type()) \
            if isinstance(l, types) \
            else base_type((l,))


# moved from datalad core, made python 3 specific
def get_func_kwargs_doc(func):
    """ Provides args for a function

    Parameters
    ----------
    func:
      function from which args are being requested

    Returns
    -------
    list
      of the args that a function takes in
    """
    return inspect.getfullargspec(func)[0]


"""DataLad crawler extension"""

__docformat__ = 'restructuredtext'

from .version import __version__

# defines a datalad command suite
# this symbold must be identified as a setuptools entrypoint
# to be found by datalad
command_suite = (
    # description of the command suite, displayed in cmdline help
    "Crawl web resources",
    [
        # specification of a command, any number of commands can be defined
        (
            # importable module that contains the command implementation
            'datalad_crawler.crawl',
            # name of the command class implementation in above module
            'Crawl',
            'crawl',
        ),
        (
            # importable module that contains the command implementation
            'datalad_crawler.crawl_init',
            # name of the command class implementation in above module
            'CrawlInit',
            'crawl-init',
            'crawl_init',
        ),
    ]
)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from datalad.support.annexrepo import AnnexRepo

from datalad.utils import chpwd
from datalad.utils import swallow_logs
from datalad.utils import make_tempfile

from logging import getLogger
lgr = getLogger('datalad.crawl.tests')


# This function needs to be in a file whose name starts with "test_" in order
# for pytest to do assertion rewriting on it.
def _test_smoke_pipelines(func, args, kwargs={}):
    with make_tempfile(mkdir=True) as tmpdir:
        AnnexRepo(tmpdir, create=True)
        with chpwd(tmpdir):
            with swallow_logs():
                for p in [func(*args, **kwargs)]:
                    assert len(p) > 1



import datalad.tests.utils_pytest
from datalad.conftest import setup_package  # autoused fixture
import pytest

_datalad_default_branch = datalad.tests.utils_pytest.DEFAULT_BRANCH

@pytest.fixture(autouse=True, scope="session")
def setup_branch():
    # Q&D workaround for not messing for now with datalad specifying default
    # branch to be not master
    datalad.tests.utils_pytest.DEFAULT_BRANCH = 'master'
    yield
    datalad.tests.utils_pytest.DEFAULT_BRANCH = _datalad_default_branch



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from os.path import join as opj

from datalad_crawler.pipelines.tests.test_utils import _test_smoke_pipelines
from ...nodes.annex import (
    Annexificator,
    initiate_dataset,
)
from datalad.dochelpers import exc_str
from datalad.utils import chpwd
from datalad.utils import _path_
from datalad.support import path as op
from datalad.support.exceptions import MissingExternalDependency
from datalad.tests.utils_pytest import with_tree
from datalad.tests.utils_pytest import eq_, assert_in
from datalad.tests.utils_pytest import with_tempfile
from datalad.tests.utils_pytest import serve_path_via_http
from datalad.tests.utils_pytest import ok_file_has_content
from datalad.tests.utils_pytest import ok_file_under_git, ok_clean_git
from datalad.tests.utils_pytest import usecase
from datalad.tests.utils_pytest import swallow_logs
from ..simple_with_archives import pipeline
from datalad.api import create

from datalad.api import crawl, crawl_init

import pytest

import logging
from logging import getLogger
lgr = getLogger('datalad.crawl.tests')


def test_smoke_pipelines():
    _test_smoke_pipelines(pipeline, ["random_url"])

from .test_balsa import TEST_TREE1

# A little integration test

@with_tree(tree=TEST_TREE1, archives_leading_dir=False)
@serve_path_via_http
@with_tempfile
def test_simple1(ind=None, topurl=None, outd=None):

    list(initiate_dataset(
        template="simple_with_archives",
        dataset_name='test1',
        path=outd,
        add_fields={'url': topurl + 'study/show/WG33',
                    'a_href_match_': '.*download.*'}
    )({}))

    with chpwd(outd):
        out, stats = crawl()

    eq_(stats.add_annex, 3)

    ok_file_under_git(outd, 'file1.nii', annexed=True)
    ok_file_has_content(opj(outd, 'file1.nii'), 'content of file1.nii')

    ok_file_under_git(outd, _path_('dir1/file2.nii'), annexed=True)
    ok_file_has_content(opj(outd, 'dir1', 'file2.nii'), 'content of file2.nii')

    eq_(len(out), 1)


TEST_TREE2 = {
    '1.tar.gz': {
        'd': {"textfile": "1\n",
              "tooshort": "1"
              },
        "anothertext": "1 2 3",
    },
}
from datalad.support.external_versions import external_versions
import inspect
if (external_versions['datalad'] >= '0.11.2'
    # for the development state
    or 'decompress' in inspect.getargspec(ok_file_has_content)[0]):
    TEST_TREE2.update({
        # Also test handling of .gz files
        # Cannot be within a subdir without adding "matchers" to go into it
        # 'sd': {
        "compressed.dat.gz": u"мама мыла раму",
        # }
    })


@pytest.mark.parametrize("gz", [
    False,
    pytest.param(
        True,
        marks=pytest.mark.skipif(
            'compressed.dat.gz' not in TEST_TREE2,
            reason="datalad 0.11.2+ required",
        )
    ),
])
@usecase  # created with
@with_tree(tree=TEST_TREE2, archives_leading_dir=False)
@serve_path_via_http
@with_tempfile
def test_crawl_autoaddtext(ind=None, topurl=None, outd=None, *, gz):
    ds = create(outd)
    ds.run_procedure("cfg_text2git")
    with chpwd(outd):  # TODO -- dataset argument
        template_kwargs = {
            'url': topurl,
            'a_href_match_': '.*',
        }
        if gz:
            template_kwargs['archives_re'] = r"\.gz$"
        crawl_init(
            template_kwargs
            , save=True
            , template='simple_with_archives'
        )
        try:
            crawl()
        except MissingExternalDependency as exc:
            pytest.skip(exc_str(exc))
    ok_clean_git(outd)
    ok_file_under_git(outd, "anothertext", annexed=False)
    ok_file_under_git(outd, "d/textfile", annexed=False)
    ok_file_under_git(outd, "d/tooshort", annexed=True)

    if 'compressed.dat.gz' in TEST_TREE2:
        if gz:
            ok_file_under_git(outd, "compressed.dat", annexed=False)
            ok_file_has_content(op.join(outd, "compressed.dat"), u"мама мыла раму")
        else:
            ok_file_under_git(outd, "compressed.dat.gz", annexed=True)
    else:
        pytest.skip("Need datalad >= 0.11.2 to test .gz files decompression")


@with_tempfile(mkdir=True)
def test_warning_no_annex_but_incoming_pipeline(outd=None):
    with chpwd(outd):  # TODO -- dataset argument
        annex = Annexificator()
        incoming_pipeline = [[[annex]]]  # deep deep inside
        with swallow_logs(new_level=logging.WARNING) as cml:
            assert pipeline(incoming_pipeline=incoming_pipeline, annex=annex)
            assert not cml.out  # we did provide annex
            assert pipeline(incoming_pipeline=incoming_pipeline)
            assert_in("incoming_pipeline already contains annexificator", cml.out)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""constants for datalad_crawler
"""

from os.path import join

from datalad.consts import DATALAD_DOTDIR
from datalad.consts import ARCHIVES_SPECIAL_REMOTE
from datalad.consts import DATALAD_SPECIAL_REMOTE

HANDLE_META_DIR = DATALAD_DOTDIR

# directory containing prepared metadata of a dataset repository:
CRAWLER_META_DIR = join(HANDLE_META_DIR, 'crawl')
CRAWLER_META_CONFIG_FILENAME = 'crawl.cfg'
CRAWLER_META_CONFIG_PATH = join(CRAWLER_META_DIR, CRAWLER_META_CONFIG_FILENAME)
CRAWLER_META_VERSIONS_DIR = join(CRAWLER_META_DIR, 'versions')
# TODO: RENAME THIS UGLINESS?
CRAWLER_META_STATUSES_DIR = join(CRAWLER_META_DIR, 'statuses')



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import os
from glob import glob
from os.path import join as opj
from os.path import exists
from mock import patch
from functools import wraps

from ...nodes.crawl_url import crawl_url
from ...nodes.matches import *
from ...pipeline import run_pipeline, FinishPipeline

from ...nodes.misc import Sink, assign, range_node, interrupt_if
from ...nodes.annex import Annexificator, initiate_dataset
from ...pipeline import load_pipeline_from_module

from datalad.support.stats import ActivityStats
from datalad.support.gitrepo import GitRepo
from datalad.support.annexrepo import AnnexRepo

from datalad.api import clean
from datalad.utils import chpwd
from datalad.utils import find_files
from datalad.utils import swallow_logs
from datalad.tests.utils_pytest import with_tree
from datalad.tests.utils_pytest import eq_, assert_not_equal, ok_, ok_startswith
from datalad.tests.utils_pytest import assert_in, assert_is_generator
from datalad.tests.utils_pytest import skip_if_no_module
from datalad.tests.utils_pytest import with_tempfile
from datalad.tests.utils_pytest import serve_path_via_http
from datalad.tests.utils_pytest import skip_if_no_network
from datalad.tests.utils_pytest import use_cassette
from datalad.tests.utils_pytest import ok_file_has_content
from datalad.tests.utils_pytest import ok_file_under_git
from datalad.downloaders.tests.utils import get_test_providers

import pytest

import logging
from logging import getLogger
lgr = getLogger('datalad.crawl.tests')


#
# Some helpers
#

from ..xnat import XNATServer
from ..xnat import PROJECT_ACCESS_TYPES
from ..xnat import superdataset_pipeline
from ..xnat import pipeline
from datalad.tests.utils_pytest import skip_if_no_network

NITRC_IR = 'https://www.nitrc.org/ir'
CENTRAL_XNAT = 'https://central.xnat.org'


def skip_xnat(func):
    """Skips test unless DATALAD_TEST_XNAT is set"""
    @wraps(func)
    def newfunc(*args, **kwargs):
        skip_if_no_network()
        if not os.getenv('DATALAD_TEST_XNAT'):
            pytest.skip('Run this test by setting DATALAD_TEST_XNAT')
        return func(*args, **kwargs)
    return newfunc


@skip_xnat
@pytest.mark.parametrize("url,project,empty_project,subjects", [
    (NITRC_IR, 'fcon_1000', None, ['xnat_S00401', 'xnat_S00447']),
    (CENTRAL_XNAT, 'CENTRAL_OASIS_LONG', 'ADHD200', ['OAS2_0001', 'OAS2_0176']),
])
@use_cassette('test_basic_xnat_interface')
def test_basic_xnat_interface(url, project, empty_project, subjects):
    nitrc = XNATServer(url)
    projects = nitrc.get_projects()
    # verify that we still have projects we want!
    assert_in(project, projects)
    if empty_project:
        all_projects = nitrc.get_projects(drop_empty=False)
        assert len(all_projects) > len(projects)
        assert empty_project in all_projects
        assert empty_project not in projects
    projects_public = nitrc.get_projects(limit='public')
    import json
    print(json.dumps(projects_public, indent=2))
    assert len(projects_public) <= len(projects)
    assert not set(projects_public).difference(projects)
    eq_(set(projects),
        set(nitrc.get_projects(limit=PROJECT_ACCESS_TYPES)))

    subjects_ = nitrc.get_subjects(project)
    assert len(subjects_)
    experiments = nitrc.get_experiments(project, subjects[0])
    # NOTE: assumption that there is only one experiment
    files1 = nitrc.get_files(project, subjects[0], experiments.keys()[0])
    assert files1

    experiments = nitrc.get_experiments(project, subjects[1])
    files2 = nitrc.get_files(project, subjects[1], experiments.keys()[0])
    assert files2

    ok_startswith(files1[0]['uri'], '/data')
    gen = nitrc.get_all_files_for_project(project,
                                          subjects=subjects,
                                          experiments=[experiments.keys()[0]])
    assert_is_generator(gen)
    all_files = list(gen)
    if len(experiments) == 1:
        eq_(len(all_files), len(files1) + len(files2))
    else:
        # there should be more files due to multiple experiments which we didn't actually check
        assert len(all_files) > len(files1) + len(files2)

@skip_xnat
@with_tempfile(mkdir=True)
def test_smoke_pipelines(d=None):
    # Just to verify that we can correctly establish the pipelines
    AnnexRepo(d, create=True)
    with chpwd(d):
        with swallow_logs():
            for p in [superdataset_pipeline(NITRC_IR)]:
                print(p)
                ok_(len(p) > 1)


@skip_xnat
@with_tempfile(mkdir=True)
def test_nitrc_superpipeline(outd=None):
    with chpwd(outd):
        pipeline = superdataset_pipeline(NITRC_IR)
        out = run_pipeline(pipeline)
    eq_(len(out), 1)
    # TODO: actual tests on what stuff was crawled


test_nitrc_superpipeline.tags = ['integration']


@skip_xnat
@with_tempfile
def test_nitrc_pipeline(outd=None):
    get_test_providers('https://www.nitrc.org/ir/')
    from datalad.distribution.dataset import Dataset
    ds = Dataset(outd).create()
    with chpwd(outd):
        out = run_pipeline(
            pipeline(NITRC_IR, project='fcon_1000', subjects=['xnat_S00401'])
        )
    eq_(len(out), 1)


test_nitrc_superpipeline.tags = ['integration']



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Basic crawler for the web
"""

import re

from boto.s3.key import Key
from boto.s3.prefix import Prefix
from boto.s3.deletemarker import DeleteMarker
import time

from datalad.utils import updated
from datalad.dochelpers import exc_str
from datalad.support.s3 import get_key_url
from datalad.support.network import iso8601_to_epoch
from datalad.downloaders.providers import Providers
from datalad.downloaders.s3 import S3Downloader
from datalad.support.exceptions import TargetFileAbsent
from ..dbs.versions import SingleVersionDB

from logging import getLogger
lgr = getLogger('datalad.crawl.s3')


def get_version_for_key(k, fmt='0.0.%Y%m%d'):
    """Given a key return a version it identifies to be used for tagging

    Uses 0.0.YYYYMMDD by default
    """
    if isinstance(k, Prefix):
        return None
    t = iso8601_to_epoch(k.last_modified)
    # format it
    return time.strftime(fmt, time.gmtime(t))


def _strip_prefix(s, prefix):
    """A helper to strip the prefix from the string if present"""
    return s[len(prefix):] if s and s.startswith(prefix) else s


class crawl_s3(object):
    """Given a source bucket and optional prefix, generate s3:// urls for the content

    """
    def __init__(self,
                 bucket,
                 prefix=None,
                 strip_prefix=True,  # either to strip leading prefix if provided
                 url_schema='s3',
                 strategy='naive',
                 versionfx=get_version_for_key,
                 repo=None,
                 ncommits=None,
                 recursive=False,
                 versioned=True,
                 exclude=None,
                 ):
        """

        Parameters
        ----------

        bucket: str
        prefix: str, optional
          Either to remember redirects for subsequent invocations
        strip_prefix: bool, optional
          Either to strip the prefix (if given) off the target paths
        versionfx: function, optional
          If not None, to define a version from the last processed key
        repo: GitRepo, optional
          Under which to store information about latest scraped version
        strategy: {'naive', 'commit-versions'}, optional
          With `naive` strategy no commits are made if there is a deletion,
          or update event, so a single run should result in a single commit
          even though interim different "load" could be added to annex under
          the same filename
        ncommits: int or None, optional
          If specified, used as max number of commits to perform.
          ??? In principle the same effect could be achieved by a node
          raising FinishPipeline after n'th commit
        recursive: bool, optional
          Either to traverse recursively or just list elements at that level
        versioned: bool, optional
          Either to expect bucket to be versioned and demand all versions per
          prefix and generate versioned urls
        exclude: str, optional
          Regular expression to search to decide which files to exclude from
          consideration
        """
        self.bucket = bucket
        if prefix and not prefix.endswith('/'):
            lgr.warning("ATM we assume prefixes to correspond only to directories, adding /")
            prefix += "/"
        self.prefix = prefix
        self.strip_prefix = strip_prefix
        self.url_schema = url_schema
        assert(strategy in {'naive', 'commit-versions'})
        self.strategy = strategy
        self.versionfx = versionfx
        self.repo = repo
        self.ncommits = ncommits
        self.recursive = recursive
        self.versioned = versioned
        self.exclude = exclude

    def __call__(self, data):

        stats = data.get('datalad_stats', None)
        url = "s3://%s" % self.bucket
        if self.prefix:
            url += "/" + self.prefix.lstrip('/')
        providers = Providers.from_config_files()
        downloader = providers.get_provider(url).get_downloader(url)

        # bucket = provider.authenticator.authenticate(bucket_name, provider.credential)
        try:
            _ = downloader.get_status(url)  # just to authenticate and establish connection
        except TargetFileAbsent as exc:
            lgr.debug("Initial URL %s lead to not something downloader could fetch: %s", url, exc_str(exc))
            pass
        bucket = downloader.bucket
        assert(bucket is not None)

        if self.repo:
            versions_db = SingleVersionDB(self.repo)
            prev_version = versions_db.version
            if prev_version and not prev_version.get('version-id', None):
                # Situation might arise when a directory contains no files, only
                # directories which we place into subdatasets
                # see https://github.com/datalad/datalad-crawler/issues/68
                # Workaround -- start from scratch
                lgr.warning("stored version-id is empty. Crawling from the beginning")
                prev_version = None
        else:
            prev_version, versions_db = None, None

        # TODO:  we could probably use headers to limit from previously crawled last-modified
        # for now will be inefficient -- fetch all, sort, proceed
        kwargs = {} if self.recursive else {'delimiter': '/'}
        all_versions = (bucket.list_versions if self.versioned else bucket.list)(self.prefix, **kwargs)
        # Comparison becomes tricky whenever as if in our test bucket we have a collection
        # of rapid changes within the same ms, so they couldn't be sorted by last_modified, so we resolve based
        # on them being marked latest, or not being null (as could happen originally), and placing Delete after creation
        # In real life last_modified should be enough, but life can be as tough as we made it for 'testing'
        def kf(k, f):
            """Some elements, such as Prefix wouldn't have any of attributes to sort by"""
            return getattr(k, f, '')
        # So ATM it would sort Prefixes first, but that is not necessarily correct...
        # Theoretically the only way to sort Prefix'es with the rest is traverse that Prefix
        # and take latest last_modified there but it is expensive, so -- big TODO if ever ;)
        # ACTUALLY -- may be there is an API call to return sorted by last_modified, then we
        # would need only a single entry in result to determine the last_modified for the Prefix, thus TODO
        cmp = lambda k: (
            kf(k, 'last_modified'),
            k.name,
            kf(k, 'is_latest'),
            kf(k, 'version_id') != 'null',
            isinstance(k, DeleteMarker)
        )

        versions_sorted = sorted(all_versions, key=cmp)  # attrgetter('last_modified'))
        # print '\n'.join(map(str, [cmp(k) for k in versions_sorted]))

        version_fields = ['last-modified', 'name', 'version-id']
        def get_version_cmp(k):
            # this one will return action version_id so we could uniquely identify
            return kf(k, 'last_modified'), k.name, kf(k, 'version_id')

        if prev_version:
            last_modified_, name_, version_id_ = [prev_version[f] for f in version_fields]
            # roll forward until we get to the element > this
            # to not breed list copies
            for i, k in enumerate(versions_sorted):
                lm, n, vid = get_version_cmp(k)
                if lm > last_modified_:
                    start = i
                    break
                elif lm == last_modified_:
                    # go by name/version_id to be matched and then switch to the next one
                    if (n, vid) == (name_, version_id_):
                        start = i+1  # from the next one
                        if stats:
                            stats.increment('skipped')
                        break
                stats.increment('skipped')
            versions_sorted = versions_sorted[start:]

        # a set of items which we have already seen/yielded so hitting any of them again
        # would mean conflict/versioning is necessary since two actions came for the same item
        staged = set()
        strategy = self.strategy
        e_prev = None
        ncommits = self.ncommits or 0

        # adding None so we could deal with the last commit within the loop without duplicating
        # logic later outside
        def update_versiondb(e, force=False):
            # this way we could recover easier after a crash
            # TODO: config crawl.crawl_s3.versiondb.saveaftereach=True
            if e is not None and (force or True):
                versions_db.version = dict(zip(version_fields, get_version_cmp(e)))
        for e in versions_sorted + [None]:
            filename = e.name if e is not None else None
            if (self.strip_prefix and self.prefix):
                filename = _strip_prefix(filename, self.prefix)
            if filename and self.exclude and re.search(self.exclude, filename):
                stats.skipped += 1
                continue

            if filename in staged or e is None:
                # we should finish this one and commit
                if staged:
                    if self.versionfx and e_prev is not None:
                        version = self.versionfx(e_prev)
                        if version is not None and version not in stats.versions:
                            stats.versions.append(version)
                    if versions_db:
                        # save current "version" DB so we would know where to pick up from
                        # upon next rerun.  Record should contain
                        # last_modified, name, versionid
                        # TODO?  what if e_prev was a DeleteMarker???
                        update_versiondb(e_prev, force=True)
                    if strategy == 'commit-versions':
                        yield updated(data, {'datalad_action': 'commit'})
                        if self.ncommits:
                            ncommits += 1
                            if self.ncommits <= ncommits:
                                lgr.debug("Interrupting on %dth commit since asked to do %d",
                                          ncommits, self.ncommits)
                                break
                    staged.clear()
                if e is None:
                    break  # we are done
            if filename:
                # might be empty if e.g. it was the self.prefix directory removed
                staged.add(filename)
            if isinstance(e, Key):
                if e.name.endswith('/'):
                    # signals a directory for which we don't care explicitly (git doesn't -- we don't! ;) )
                    continue
                url = get_key_url(e, schema=self.url_schema, versioned=self.versioned)
                # generate and pass along the status right away since we can
                yield updated(
                    data,
                    {
                        'url': url,
                        'url_status': S3Downloader.get_key_status(e, dateformat='iso8601'),
                        'filename': filename,
                        'datalad_action': 'annex',
                    })
                update_versiondb(e)
            elif isinstance(e, DeleteMarker):
                if strategy == 'commit-versions':
                    # Since git doesn't care about empty directories for us makes sense only
                    # in the case when DeleteMarker is not pointing to the subdirectory
                    # and not empty (if original directory was removed)
                    if filename and not filename.endswith('/'):
                        yield updated(data, {'filename': filename, 'datalad_action': 'remove'})
                    else:
                        # Situation there is much trickier since it seems that "directory"
                        # could also be a key itself and created/removed which somewhat interfers with
                        # all our logic here
                        # For an interesting example see
                        #  s3://openneuro/ds000217/ds000217_R1.0.0/compressed
                        lgr.info("Ignoring DeleteMarker for %s", filename)

                update_versiondb(e)
            elif isinstance(e, Prefix):
                # so  we were provided a directory (in non-recursive traversal)
                assert(not self.recursive)
                yield updated(
                    data,
                    {
                        'url': url,
                        'filename': filename.rstrip('/'),
                        'datalad_action': 'directory',
                    }
                )
            else:
                raise ValueError("Don't know how to treat %s" % e)
            e_prev = e



from ..utils import (
    flatten,
    get_func_kwargs_doc,
)

from datalad.tests.utils_pytest import assert_equal

def test_flatten():

    assert_equal(flatten([]), [])
    assert_equal(flatten([1]), [1])
    assert_equal(flatten([1, 2]), [1, 2])
    # tuples etc aren't flattened
    assert_equal(flatten([1, {2}, (3, 4)]), [1, {2}, (3, 4)])
    # but we could make them flatten too!
    assert_equal(flatten([1, {2}, (3, 4)], types=(set, tuple, list)),
                 [1, 2, 3, 4])

    # recurse now
    assert_equal(flatten([1, [2, [3, 4], 5], 6]), [1, 2, 3, 4, 5, 6])

    # And now try "fancy" (original implementation target was list) types
    assert_equal(flatten(((0,), (1, 2))), (0, 1, 2))
    assert_equal(flatten(({0}, (1, 2)), types=(set, tuple)), (0, 1, 2))
    assert_equal(flatten([(0,), {1: 2}], types=(list, tuple, dict), base_type=tuple),
                 (0, 1))


def test_get_func_kwargs_doc():
    def some_func(arg1, kwarg1=None, kwarg2="bu"):
        return
    assert_equal(get_func_kwargs_doc(some_func), ['arg1', 'kwarg1', 'kwarg2'])



#!/usr/bin/env python

from setuptools import setup
from setuptools import find_packages
from setuptools import findall

from os.path import join as opj
from os.path import sep as pathsep
from os.path import splitext
from os.path import dirname

from setup_support import BuildManPage
from setup_support import BuildRSTExamplesFromScripts
from setup_support import get_version


def findsome(subdir, extensions):
    """Find files under subdir having specified extensions

    Leading directory (datalad) gets stripped
    """
    return [
        f.split(pathsep, 1)[1] for f in findall(opj('datalad_crawler', subdir))
        if splitext(f)[-1].lstrip('.') in extensions
    ]

# extension version
version = get_version()

cmdclass = {
    'build_manpage': BuildManPage,
    'build_examples': BuildRSTExamplesFromScripts,
}

with open(opj(dirname(__file__), 'README.md')) as fp:
    long_description = fp.read()

requires = {
    'core': [
        'datalad>=0.14.0',
        'scrapy>=1.1.0',  # versioning is primarily for python3 support
    ],
    'devel-docs': [
        # Documentation
        'sphinx',
        'sphinx-rtd-theme',
    ],
    'tests': [
        'datalad>=0.17.0',
        'pytest>=7.0',
        'pytest-cov',
        'mock',
    ],
}
requires['devel'] = sum(list(requires.values()), [])

setup(
    # basic project properties can be set arbitrarily
    name="datalad_crawler",
    author="The DataLad Team and Contributors",
    author_email="team@datalad.org",
    version=version,
    description="DataLad extension package for crawling external web resources into an automated data distribution",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=[pkg for pkg in find_packages('.') if pkg.startswith('datalad')],
    # datalad command suite specs from here
    install_requires=requires['core'],
    extras_require=requires,
    cmdclass=cmdclass,
    entry_points = {
        # 'datalad.extensions' is THE entrypoint inspected by the datalad API builders
        'datalad.extensions': [
            # the label in front of '=' is the command suite label
            # the entrypoint can point to any symbol of any name, as long it is
            # valid datalad interface specification (see demo in this extension)
            'crawler=datalad_crawler:command_suite',
        ],
        'datalad.tests': [
            'crawler=datalad_crawler',
        ],
    },
)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""A pipeline for crawling XNAT servers/datasets (e.g. NITRC/ir)"""

"""
Notes:

- xml entries for subject contain meta-data
- in longitudinal studies there could be multiple ages for a subject... find where
  it is
"""

import json
from ..nodes.misc import assign
from ..nodes.annex import Annexificator
from datalad.utils import updated

# Possibly instantiate a logger if you would like to log
# during pipeline creation
from logging import getLogger
lgr = getLogger("datalad.crawler.pipelines.xnat")

from datalad.tests.utils_pytest import eq_
from datalad.utils import assure_list, assure_bool


def list_to_dict(l, field):
    return {r.pop(field): r for r in l}


DEFAULT_RESULT_FIELDS = {'totalrecords', 'result'}
PROJECT_ACCESS_TYPES = {'public', 'protected', 'private'}


def lower_case_the_keys(d):
    """Create a dict with keys being lower cased but with a check against collisions
    
    If d is a list of tuple, would recurse into its elements
    """
    if not isinstance(d, dict):
        assert isinstance(d, (list, tuple))
        return d.__class__(
            lower_case_the_keys(x) for x in d
        )
    out = {}
    for k, v in d.items():
        kl = k.lower()
        if kl in out:
            raise ValueError(
                "We already got key %s=%r, but trying to add =%r"
                % (kl, out[kl], v)
            )
        out[kl] = v
    return out


def extract_subject_info(data):
    subject_item = data['items'][0]
    info = {'label': subject_item['data_fields']['label'], 
            'group': subject_item['data_fields'].get('group')}
    for child in subject_item['children']:
        if child['field'] == 'demographics':
            demographics = child['items'][0]['data_fields']
            info['gender'] = demographics.get('gender')
            info['age'] = demographics.get('age')
            info['handedness'] = demographics.get('handedness')
            break
    return info


def extract_experiment_info(data):
    experiment_item = data['items'][0]
    return experiment_item['data_fields']


class XNATServer(object):

    def __init__(self, topurl):
        self.topurl = topurl
        from datalad.downloaders.providers import Providers
        providers = Providers.from_config_files()
        self.downloader = providers.get_provider(topurl).get_downloader(topurl)
        # set without various methods
        self._experiment_labels = None

    @property
    def experiment_labels(self):
        if self._experiment_labels is None:
            lgr.debug("Experiment labels are not known yet")
            self.get_all_experiments()
        return self._experiment_labels

    def __call__(self, query,
                 format='json',
                 options=None,
                 return_plain=False,
                 fields_to_check=DEFAULT_RESULT_FIELDS):
        query_url = "%s/%s" % (self.topurl, query)
        options = options or {}
        if format:
            options['format'] = format
        if options:
            # TODO: use the helper we have
            query_url += "?" + '&'.join(("%s=%s" % (o, v) for o, v in options.items()))
        out = self.downloader.fetch(query_url)
        if format == 'json':
            j = json.loads(out)
            j = lower_case_the_keys(j)
            if return_plain:
                return j
            assert list(j) == ['resultset']

            jrs = lower_case_the_keys(j['resultset'])
            if fields_to_check:
                eq_(set(jrs.keys()), fields_to_check)
            return lower_case_the_keys(jrs['result'])
        return out

    def get_projects(self, limit=None, drop_empty=True, asdict=True):
        """Get list of projects 

        Parameters
        ----------
        limit: {'public', 'protected', 'private', None} or list of thereof
           'private' -- projects you have no any access to. 'protected' -- you could
           fetch description but not the data. None - would list all the projects

        drop_empty: bool
          whether to drop projects with no experiments.  Implies
          limit = 'public' since we need access to projects to determine if they
          are empty.  If drop_empty is set nd limit is None, limit is set to
          'public'
        """

        limit = assure_list(limit)

        if drop_empty:
            # TODO: rework this logic/restriction
            #  yoh thinks this should not be implied (we could check "private"
            #  ones we have access to) here.
            if limit != ['public']:
                lgr.warning(
                    "Limit was set to %s but ATM we can drop_empty only if "
                    "limit is 'public'. Reset to 'public'",
                    repr(limit)
                )
            limit = ['public']


        kw = {}

        if limit:
            # Q yoh:  is this only for public? then check above should be
            #   if 'public' in limit, and assure_list(limit) should go above
            kw['options'] = {"accessible": "true"}
            kw['fields_to_check'] = DEFAULT_RESULT_FIELDS | {'title', 'xdat_user_id'}

        all_projects = self('data/projects', **kw)

        if limit:
            # double check that all the project_access thingies in the set which
            # we know
            assert all(p['project_access'] in PROJECT_ACCESS_TYPES
                       for p in all_projects)
            limited_projects = [
                p for p in all_projects
                if p['project_access'] in limit
            ]
        else:
            limited_projects = all_projects

        experiments = self.get_all_experiments()

        if drop_empty:
            # first pass: exclude projects without experiments
            non_empty_projects = set([ e['project'] for e in experiments ])
            out = []
            # second pass: descend into experiments to find files
            for p in limited_projects:
                if self.project_has_files(p['id']):
                    out.append(p)
                else:
                    lgr.debug("Skipping %s (no files)", p)
            lgr.info("Skipped %d empty projects", len(all_projects) - len(out))
        else:
            out = limited_projects

        return list_to_dict(out, 'id') if asdict else out

    def get_all_experiments(self):
        """Query XNAT server for all known experiments.

        Side effect: (re)assigns self._experiment_labels mapping which is also
        used for crawling an individual dataset

        Returns
        -------
        list of experiments
        """
        lgr.debug("Requesting full list of experiments")
        fields_to_check = DEFAULT_RESULT_FIELDS.union({'title'})
        experiments = self('data/experiments', fields_to_check=fields_to_check)
        self._experiment_labels = {e['id']: e['label'] for e in experiments}
        return experiments

    def project_has_files(self, project):
        for subject in self.get_subjects(project):
            for experiment in self.get_experiments(project, subject):
                if self.get_files(project, subject, experiment):
                    return True
        return False

    def get_subjects(self, project):
        return list_to_dict(
            self('data/projects/%(project)s/subjects' % locals()),
            'id'
        )

    def get_experiments(self, project, subject):
        fmt = 'data/projects/%(project)s/subjects/%(subject)s/experiments'
        url = fmt % locals()
        return list_to_dict(
            self(url, options={'xsiType': 'xnat:imageSessionData'}),
            'id'
        )

    def get_files(self, project, subject, experiment):
        files = []
        for rec_type in ('scans', 'assessors'):
            url = 'data/projects/%(project)s/subjects/%(subject)s/experiments/%(experiment)s/%(rec_type)s' % \
                      locals()
            recs = self(url)
            files_ = []
            for rec in recs:
                files_.extend(self(url + '/%(id)s/files' % rec, fields_to_check={'title', 'result', 'columns'}))
            for f in files_:
                f['rec_type'] = rec_type
            files.extend(files_)
        return files

    def get_all_files_for_project(self, project, subjects=None, experiments=None):
        # TODO: grow the dictionary with all the information about subject/experiment/file
        # to be yielded so we could tune up file name anyway we like
        for subject in (subjects or self.get_subjects(project)):
            subject_url = 'data/projects/%s/subjects/%s' % (project, subject)
            subject_data = self(subject_url, return_plain=True)
            subject_info = extract_subject_info(subject_data)
            for experiment in (experiments or self.get_experiments(project, subject)):
                experiment_data = self('data/experiments/%s' % experiment, 
                                       return_plain=True)
                experiment_info = extract_experiment_info(experiment_data)
                for file_ in self.get_files(project, subject, experiment):
                    file_info = updated(file_, {'subject_id': subject, 
                                                'subject_info': subject_info, 
                                                'experiment_id': experiment, 
                                                'experiment_info': experiment_info})
                    yield file_info


# define a pipeline factory function accepting necessary keyword arguments
# Should have no strictly positional arguments
def superdataset_pipeline(url, limit=None, drop_empty=True):
    """
    
    Parameters
    ----------
    url
    limit : TODO, optional
      Types of access to limit to, see XNAT.get_datasets
    drop_empty: bool, optional
      If set, do not create datasets which are empty (no files).
      Note - it requires obtaining details for every project, which could be
      a heavy operation
    kwargs

    Returns
    -------

    """

    annex = Annexificator(no_annex=True, allow_dirty=False)
    lgr.info("Creating a pipeline with url=%s limit=%s drop_empty=%s", url, limit, drop_empty)
    limit = assure_list(limit)
    drop_empty = assure_bool(drop_empty)

    def get_projects(data):
        xnat = XNATServer(url)
        for p in xnat.get_projects(
                asdict=False,
                limit=limit or PROJECT_ACCESS_TYPES,
                drop_empty=drop_empty
        ):
            yield updated(data, p)

    return [
        get_projects,
        assign({'project': '%(id)s',
                'dataset_name': '%(id)s',
                'url': url
                }, interpolate=True),
        # TODO: should we respect  x quarantine_status
        annex.initiate_dataset(
            template="xnat",
            data_fields=['project', 'url', 'project_access'],  # TODO: may be project_access
            # let's all specs and modifications reside in master
            # branch='incoming',  # there will be archives etc
            existing='skip'
            # further any additional options
        )
    ]


def pipeline(url, project, project_access='public', subjects=None):
    # TODO: Ben: Clarify parameters. In particular `project_access` is unclear to me
    subjects = assure_list(subjects)

    xnat = XNATServer(url)

    def get_project_info(data):
        out = xnat('data/projects/%s' % project,
                   return_plain=True
                   )
        # for NITRC I need to get more!
        # "http://nitrc_es.projects.nitrc.org/datalad/%s" % dataset
        items = out['items']
        assert len(items) == 1
        dataset_meta = items[0]['data_fields']
        # TODO: save into a file
        yield data

    def get_files(data):

        for f in xnat.get_all_files_for_project(project, subjects=subjects):
            # TODO: tune up filename
            # TODO: get url
            prefix = '/data/experiments/'
            assert f['uri'].startswith('%s' % prefix)
            # TODO:  use label for subject/experiment
            # TODO: might want to allow for
            #   XNAT2BIDS whenever that one is available:
            #     http://reproducibility.stanford.edu/accepted-projects-for-the-2nd-crn-coding-sprint/
            exp_label = xnat.experiment_labels[f['experiment_id']]
            yield updated(data,
                          {'url': url + f['uri'],
                           'path': f['uri'][len(prefix):], 
                           'name': '%s-%s' % (exp_label, f['name'])
                           })

    annex = Annexificator(
        create=False,  # must be already initialized etc
        # leave in Git only obvious descriptors and code snippets -- the rest goes to annex
        # so may be eventually we could take advantage of git tags for changing layout
        statusdb='json',
        special_remotes=['datalad'] if project_access != 'public' else None
    )

    return [
        get_project_info,
        [
            get_files,
            annex
        ],
        annex.finalize(cleanup=True, aggregate=True),
    ]



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""File-based "DB" which uses file modification times to deduce if new version is available

"""

import os
from os.path import exists, lexists, islink, realpath, sep

from datalad.dochelpers import exc_str
from datalad.support.status import FileStatus
from datalad.support.exceptions import CommandError
from datalad.utils import auto_repr
from datalad.utils import disable_logger
from datalad_crawler.consts import CRAWLER_META_STATUSES_DIR

from .base import JsonBaseDB, FileStatusesBaseDB
import logging
lgr = logging.getLogger('datalad.crawler.dbs')

__docformat__ = 'restructuredtext'

__all__ = ['PhysicalFileStatusesDB']

#
# Concrete implementations
#

@auto_repr
class PhysicalFileStatusesDB(FileStatusesBaseDB):
    """Non-persistent DB based on file attributes

    It uses file modification times and size as known to annex to deduce
    if a new version is available

    In general, this should not be used since neither git nor annex stores any files
    meta-information (besides mode and size), so mtime would get lost while switching
    the branches and dropping the load
    """

    def _get(self, filepath):
        if not lexists(filepath):
            return None

        # I wish I could just test using filesystem stats but that would not
        # be reliable, and also file might not even be here.
        # File might be under git, not annex so then we would need to assess size
        filestat = os.lstat(filepath)
        try:
            with disable_logger():
                info = self.annex.info(filepath, batch=True)
            size = info['size']
        except (CommandError, TypeError) as exc:
            # must be under git or a plain file
            lgr.debug("File %s must be not under annex, since info failed: %s" % (filepath, exc_str(exc)))
            size = filestat.st_size

        # deduce mtime from the file or a content which it points to. Take the oldest (I wonder
        # if it would bite ;) XXX)
        mtime = filestat.st_mtime
        if islink(filepath):
            filepath_ = realpath(filepath)  # symlinked to
            if exists(filepath_):
                mtime_ = os.stat(filepath_).st_mtime
                mtime = min(mtime_, mtime)
        return FileStatus(
                size=size,
                mtime=mtime
        )

    def _set(self, filepath, status):
        # this DB doesn't implement much of it, besides marking internally that we do care about this file
        pass

    def save(self):
        # nothing for us to do but JsonFileStatusesDB has it so let's keep it common
        pass

    def _remove(self, filepath):
        pass


@auto_repr
class JsonFileStatusesDB(JsonBaseDB, PhysicalFileStatusesDB):
    """Persistent DB to store information about files' size/mtime/filename in a JSON file
    """

    __version__ = 1
    __crawler_subdir__ = CRAWLER_META_STATUSES_DIR

    def __init__(self, annex, track_queried=True, name=None):
        PhysicalFileStatusesDB.__init__(self, annex, track_queried=track_queried)
        JsonBaseDB.__init__(self, annex, name=name)

    #
    # Defining abstract methods implementations
    #
    def _get_empty_db(self):
        return {'files': {}}

    def _get_loaded_db(self, db):
        """Given a DB loaded from a file, prepare it for being used
        """
        assert (set(db.keys()) == {'db_version', 'files'})
        return db

    def _get_db_to_save(self):
        """Return DB to be saved as JSON file
        """
        return self._db

    def _get_fpath(self, filepath):
        assert (filepath.startswith(self.annex.path))
        fpath = filepath[len(self.annex.path.rstrip(sep)) + 1:]
        return fpath

    def _get_fileattributes_status(self, fpath):
        filepath = self._get_filepath(fpath)
        return PhysicalFileStatusesDB._get(self, filepath)

    def _get(self, filepath):
        # TODO: may be avoid this all fpath -> filepath -> fpath?
        fpath = self._get_fpath(filepath)

        files = self._db['files']
        if fpath not in files:
            return None
        return FileStatus(**files[fpath])

    # TODO: get URL all the way here?
    def _set(self, filepath, status):
        fpath = self._get_fpath(filepath)
        if status is None:
            status_dict = {}
            # get it from the locally available file
            # status = PhysicalFileStatusesDB._get(self, filepath)
            # NOPE since then generated files would keep changing
        else:
            status_dict = {f: getattr(status, f)
                           for f in ('size', 'mtime', 'filename')
                           if getattr(status, f) is not None}
        self._db['files'][fpath] = status_dict

    def _remove(self, filepath):
        fpath = self._get_fpath(filepath)
        if fpath in self._db['files']:
            self._db['files'].pop(fpath)
